#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <values.h>
#include <SDL.h>

#include "barcode.h"
#include "image.h"
#include "video.h"

static void set_create(Set *set, size_t size) {
    set->n = 0;
    set->alloced = 16;
    set->size = size;
    set->data = malloc(16*size);
}

static void set_add(Set *set, const void *data) {
    if(set->n >= set->alloced) {
        set->alloced *= 2;
        set->data = realloc(set->data, set->alloced*set->size);
    }
    memcpy(set->data + (set->size*set->n), data, set->size);
    set->n++;
}

static void set_delete(Set *set, int i) {
    if(i < set->n-1) {
        memcpy(set->data + (i*set->size),
            set->data + ((set->n-1)*set->size), set->size);
    }
    set->n--;
}

static void grey_scan_line(Grey *grey, const float *p, int y, Set *bars) {
    float pmax, scale, q, last_p;
    int i,j,last_i,n=0,first=8,ratio,added;
    int x[grey->x];
    const int bits = CODE_BITS;
    int len[bits];
    int ones,zeros,one_sum,zero_sum,diff_sum,tres,zero_norm,one_norm;
    int word;
    Bar bar, *old;

    pmax = p[first];
    for(i=first; i<grey->x; i++)
        if(p[i] > pmax) pmax = p[i];
    scale = 1.0f / pmax;
    last_p = p[first]*scale;
    last_i = first;
    for(i=first+1; i<grey->x; i++) {
        q = p[i]*scale;
        if(q > last_p) {
            last_p = q;
            last_i = i;
        } else if(i > last_i + LINE_LOCAL) {
            if(last_p > LINE_TRES)
                x[n++] = last_i;
            last_i = i;
            last_p = q;
        }
    }
    for(i=0; i<n-bits; i++) {
        len[0] = x[i+1] - x[i];
        len[bits-1] = x[i+bits] - x[i+bits-1];
        ratio = (0x1000 * len[bits-1]) / len[0];
        if(ratio >= 0x1000-0x1000/LINE_TOL &&
           ratio <= 0x1000+0x1000/LINE_TOL) {
            ones = 0;
            one_sum = 0;
            zeros = 2;
            zero_sum = len[0] + len[bits-1];
        } else if(ratio >= 0x2000-0x2000/LINE_TOL &&
                  ratio <= 0x2000+0x2000/LINE_TOL) {
            ones = 1;
            one_sum = len[bits-1];
            zeros = 1;
            zero_sum = len[0];
        } else continue;
        for(j=1; j<bits-1; j++) {
            len[j] = x[i+j+1] - x[i+j];
            if(zeros*len[j] > zero_sum+zero_sum/2) {
                ones++;
                one_sum += len[j];
            } else {
                zeros++;
                zero_sum += len[j];
            }
        }
        if(ones != zeros-1) continue;
        word = 0;
        zero_norm = zero_sum/zeros;
        one_norm = (ones)? one_sum/ones : zero_norm*2;
        tres = (zero_norm + one_norm) / 2;
        diff_sum = 0;
        for(j=1; j<bits-1; j++) {
            if(len[j] < tres) {
                diff_sum += abs(len[j] - zero_norm);
                word *= 2;
            } else {
                diff_sum += abs(len[j] - one_norm);
                word = 1 + word*2;
            }
        }
        if(diff_sum <= zero_norm*(ones+zeros)/LINE_TOL) {
            bar.x0 = x[i];
            bar.x1 = x[i+bits];
            bar.y0 = y;
            bar.y1 = y;
            bar.n = 1;
            /* fprintf(stderr, "%d-%dx%d\n",bar.x0,bar.x1,bar.y0); */
            bar.word = word;
            added = 0;
            for(j=0; j<bars->n; j++) {
                old = ((Bar*)(bars->data)) + j;
                if(old->word != word) continue;
                if(abs(old->x0-bar.x0) + abs(old->x1-bar.x1) > 32) continue;
                if(bar.y0 - old->y1 > 32) continue;
                old->x0 = bar.x0;
                old->x1 = bar.x1;
                old->y1 = y;
                old->n++;
                added = 1;
                break;
            }
            if(!added) set_add(bars, &bar);
        }
    }
}

static int compar_bars(const void *a, const void *b) {
    const Bar *bar0 = a;
    const Bar *bar1 = b;
    return ((bar1->y0+bar1->y1/2) - (bar0->y0+bar0->y1/2));
}

static Set *grey_scan(Grey *grey) {
    int i,j;
    float sum[grey->x];
    Set *bars;
    Bar *bar;

    bars = malloc(sizeof(*bars));
    set_create(bars, sizeof(Bar));
    for(i=0; i<grey->x; i++) {
        sum[i] = grey->pixels[i] * (float)LINE_AVG;
    }
    for(j=1; j<grey->y; j++) {
        for(i=0; i<grey->x; i++) {
            if(j <= LINE_AVG) {
                sum[i] -= grey->pixels[i];
            } else {
                sum[i] -= grey->pixels[(j-LINE_AVG)*grey->x + i];
            }
            sum[i] += grey->pixels[j*grey->x + i];
        }
        grey_scan_line(grey, sum, j, bars);
    }
    for(i=0; i<bars->n; i++) {
        bar = ((Bar*)(bars->data)) + i;
        if(bar->n < LINE_NR) {
            set_delete(bars, i);
            i--;
        }
    }
    if(bars->n == 0) {
        free(bars->data);
        free(bars);
        return NULL;
    }
    qsort(bars->data, bars->n, sizeof(Bar), compar_bars);
    return bars;
}

static void grey_draw(Grey *g1, Grey *g2, Grey *g3, SDL_Surface *surface) {
    int i;
    Uint32 *p, c1, c2, c3;
    int x,y;
    if(g1 != NULL) { x = g1->x; y = g1->y; }
    else if(g2 != NULL) { x = g2->x; y = g2->y; }
    else if(g3 != NULL) { x = g3->x; y = g3->y; }
    p = surface->pixels;
    for(i=0; i<x*y; i++) {
        c1 = c2 = c3 = 0;
        if(g1 != NULL) c1 = (int)(255.0f*g1->pixels[i]);
        if(g2 != NULL) c2 = (int)(255.0f*g2->pixels[i]);
        if(g3 != NULL) c3 = (int)(255.0f*g3->pixels[i]);
        *(p++) = SDL_MapRGB(surface->format, c1, c2, c3);
    }
}

static int detect_code(Set *bars, int *buf, int rows) {
    int i,j,x0,x1,word;
    Bar *bar;
    if(bars == NULL) return 0;
    for(i=0; i<=bars->n-rows; i++) {
        for(j=0; j<rows; j++) {
            bar = ((Bar*)(bars->data)) + (i+j);
            if(j > 0) {
                if(abs(x0-bar->x0) + abs(x1-bar->x1) > 32) break;
                if(bar->word == word) break;
            }
            x0 = bar->x0;
            x1 = bar->x1;
            word = bar->word;
            buf[j] = word;
        }
        if(j == rows) return 1;
    }
    return 0;
}

static int video_scan(Video *video) {
    Grey *frame, *soft, *diff;
    Set *bars;
    SDL_Surface *screen;
    Uint8 *keys;
    int data[CODE_ROWS];
    int i;

    if(SDL_Init(SDL_INIT_VIDEO) < 0) {
        fputs(SDL_GetError(),stderr);
        return 1;
    }
    atexit(SDL_Quit);
    if(!(screen = SDL_SetVideoMode(video->x,video->y,32,SDL_SWSURFACE))) {
        fputs(SDL_GetError(),stderr);
        return 1;
    }
    do {
        SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
        frame = video_grey_snapshot(video);
#if SMOOTHING
        soft = grey_convolute(frame);
        free(frame);
#else
        soft = frame;
#endif
        diff = grey_diff(soft);
        bars = grey_scan(diff);
        grey_normalize(diff);
        if(bars != NULL) {
            for(i=0; i<bars->n; i++) {
                Bar *bar;
                bar = ((Bar*)(bars->data)) + i;
                grey_rect(soft, bar->x0, bar->y0, bar->x1, bar->y1, 1.0f);
            }
        }
        if(detect_code(bars, data, CODE_ROWS)) {
            printf("%x", data[0]);
            for(i=1; i<CODE_ROWS; i++)
                printf(" %x", data[i]);
            putchar('\n');
        }
        grey_draw(soft, soft, diff, screen);
        free(soft);
        free(diff);
        SDL_Flip(screen);
        SDL_PumpEvents();
        keys = SDL_GetKeyState(NULL);
    } while(!keys[SDLK_ESCAPE]);
    return 0;
}

int main(int argc, char *argv[]) {
    char *filename;
    Video video;

    if(argc > 1) filename = argv[1]; else filename = "/dev/video0";
    if(video_init(&video, filename, 640, 480)) {
        fprintf(stderr, "Error initializing video.\n");
        return 1;
    }
    video_scan(&video);
    video_close(&video);
    return 0;
}

