#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "barcode.h"
#include "image.h"

static int read_int(FILE *file, int *x) {
    int c;
    for(;;) {
        do { c = getc(file); } while(isspace(c));
        if(c == EOF) return 1;
        if(c != '#') {
            ungetc(c,file);
            break;
        }
        for(;;) {
            c = getc(file);
            if(c == '\n') break;
            if(c == EOF) return 1;
        }
    }
    if(fscanf(file, "%d", x) != 1) return 1;
    return 0;
}

PPM *ppm_load(FILE *file) {
    PPM *ppm;
    int x,y,max;
    if(getc(file) != 'P') return NULL;
    if(getc(file) != '6') return NULL;
    if(read_int(file, &x)) return NULL;
    if(read_int(file, &y)) return NULL;
    if(read_int(file, &max)) return NULL;
    if(x > 0x1000 || y > 0x1000) return NULL;
    ppm = malloc(sizeof(*ppm)+sizeof(unsigned char)*3*x*y);
    ppm->x = x;
    ppm->y = y;
    ppm->max = max;
    if(fread(ppm->pixels, 3*sizeof(unsigned char), x*y, file) != x*y) {
        free(ppm);
        return NULL;
    }
    return ppm;
}

void ppm_write(PPM *ppm, FILE *file) {
    fprintf(file, "P6\n%d %d\n%d\n", ppm->x, ppm->y, ppm->max);
    fwrite(ppm->pixels, 3*sizeof(unsigned char), ppm->x*ppm->y, file);
}

Grey *ppm_to_grey(PPM *ppm) {
    Grey *grey;
    int i,j;

    grey = malloc(sizeof(*grey)+ppm->x*ppm->y*sizeof(*(grey->pixels)));
    grey->x = ppm->x;
    grey->y = ppm->y;
    for(i=j=0; i<grey->x*grey->y; i++,j+=3) {
        grey->pixels[i] = ((float)
            (ppm->pixels[j+0]+ppm->pixels[j+1]+ppm->pixels[j+2])) *
            (1.0f/765.0f);
    }
    return grey;
}

PPM *grey_to_ppm(Grey *grey) {
    int i,j;
    unsigned char c;
    PPM *ppm;
    ppm = malloc(sizeof(*ppm)+sizeof(unsigned char)*3*grey->x*grey->y);
    ppm->x = grey->x;
    ppm->y = grey->y;
    ppm->max = 255;
    for(i=0,j=0; i<grey->x*grey->y; i++) {
        c = (unsigned char)(255.0f*grey->pixels[i]);
        ppm->pixels[j++] = c;
        ppm->pixels[j++] = c;
        ppm->pixels[j++] = c;
    }
    return ppm;
}

static float grey_pixel(Grey *grey, int x, int y) {
    if(x < 0) x = 0;
    if(x >= grey->x) x = grey->x - 1;
    if(y < 0) y = 0;
    if(y >= grey->y) y = grey->y - 1;
    return grey->pixels[y*grey->x + x];
}

static Grey *grey_copy(Grey *grey) {
    Grey *t;
    t = malloc(sizeof(*grey)+grey->x*grey->y*sizeof(*(grey->pixels)));
    t->x = grey->x;
    t->y = grey->y;
    return t;
}

Grey *grey_convolute(Grey *grey) {
    Grey *conv;
    int a,b,i,j,k,p,q;
    float s;
    const int m = 5, n = 5;
    const float A[5][5] = {
        {   2.0/159.0,  4.0/159.0,  5.0/159.0,  4.0/159.0,  2.0/159.0   },
        {   4.0/159.0,  9.0/159.0, 12.0/159.0,  9.0/159.0,  4.0/159.0   },
        {   5.0/159.0, 12.0/159.0, 15.0/159.0, 12.0/159.0,  5.0/159.0   },
        {   4.0/159.0,  9.0/159.0, 12.0/159.0,  9.0/159.0,  4.0/159.0   },
        {   2.0/159.0,  4.0/159.0,  5.0/159.0,  4.0/159.0,  2.0/159.0   }
    };

    conv = grey_copy(grey);
    k = 0;
    for(j=0; j<grey->y; j++) {
        for(i=0; i<grey->x; i++) {
            s = 0.0f;
            for(q=j-(m-1)/2,b=0; b<m; q++,b++) {
                for(p=i-(n-1)/2,a=0; a<n; p++,a++)
                    s += grey_pixel(grey,p,q) * A[b][a];
            }
            conv->pixels[k] = s;
            k++;
        }
    }
    return conv;
}

Grey *grey_diff(Grey *grey) {
    Grey *diff;
    int i,j,k;
    float dx,dy;

    diff = grey_copy(grey);
    k = 0;
    for(j=0; j<grey->y; j++) {
        for(i=0; i<grey->x; i++) {
            dx = grey_pixel(grey,i+1,j)-grey_pixel(grey,i-1,j);
            dy = grey_pixel(grey,i,j+1)-grey_pixel(grey,i,j-1);
            dx = fabsf(dx);
            dy = fabsf(dy);
            diff->pixels[k++] = dx;
        }
    }
    return diff;
}

void grey_normalize(Grey *grey) {
    int i;
    float maxp = 0.0f, scale;
    for(i=0; i<grey->x*grey->y; i++) {
        if(grey->pixels[i] > maxp) maxp = grey->pixels[i];
    }
    scale = (maxp > 0.0f)? 1.0f / maxp : 0.0f;
    for(i=0; i<grey->x*grey->y; i++) {
        grey->pixels[i] *= scale;
    }
}

void grey_rect(Grey *grey, int x0, int y0, int x1, int y1, float c) {
    int i;
    for(i=x0; i<=x1; i++) {
        grey->pixels[y0*grey->x + i] = c;
        grey->pixels[y1*grey->x + i] = c;
    }
    for(i=y0+1; i<=y1-1; i++) {
        grey->pixels[i*grey->x + x0] = c;
        grey->pixels[i*grey->x + x1] = c;
    }
}

