#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <asm/types.h>
#include <linux/videodev2.h>

#include "barcode.h"
#include "image.h"
#include "video.h"

/* TODO: mmap */
/* http://v4l2spec.bytesex.org/spec/capture-example.html
 * http://v4l2spec.bytesex.org/spec/book1.htm
 */
int video_init(Video *video, const char *filename, int x, int y) {
    struct v4l2_format fmt;
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;
    struct v4l2_requestbuffers req;
    struct v4l2_capability cap;
    enum v4l2_buf_type type;
    int i;
    int fd;

    if((fd = open(filename, O_RDWR|O_NONBLOCK, 0)) < 0) {
        fprintf(stderr, "Error opening video device: %s\n", filename);
        return 1;
    }
    if(ioctl(fd, VIDIOC_QUERYCAP, &cap) < 0) {
        fprintf(stderr,
        "VIDIOC_QUERYCAP ioctl failed (is %s a video device?)\n", filename);
        return 1;
    }
    if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        fprintf(stderr, "Video capture not supported.\n");
        return 1;
    }

    memset(&cropcap, 0, sizeof(cropcap));
    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(!ioctl(video->fd, VIDIOC_CROPCAP, &cropcap)) {
        crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        crop.c = cropcap.defrect;
        ioctl(video->fd, VIDIOC_S_CROP, &crop);
    }
    memset(&fmt, 0, sizeof(fmt));
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width = x;
    fmt.fmt.pix.height = y;
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
    if(ioctl(fd, VIDIOC_S_FMT, &fmt) < 0) {
        fprintf(stderr, "VIDIOC_S_FMT ioctl failed.\n");
        return 1;
    }
    video->x = fmt.fmt.pix.width;
    video->y = fmt.fmt.pix.height;
    video->fd = fd;
    req.count = 4;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;
    if(ioctl(video->fd, VIDIOC_REQBUFS, &req) < 0) {
        fprintf(stderr, "Memory mapping not supported.\n");
        return 1;
    }
    video->buffers = calloc(req.count, sizeof(*(video->buffers)));
    video->nbuffers = req.count;
    for(i=0; i<req.count; i++) {
        struct v4l2_buffer buf;
        memset(&buf, 0, sizeof(buf));
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = i;
        if(ioctl(video->fd, VIDIOC_QUERYBUF, &buf) < 0) {
            fprintf(stderr, "VIDIOC_QUERYBUF ioctl failed.\n");
            return 1;
        }
        video->buffers[i].length = buf.length;
        video->buffers[i].start =
            mmap(NULL, buf.length, PROT_READ|PROT_WRITE,
                 MAP_SHARED, video->fd, buf.m.offset);
        if(video->buffers[i].start == MAP_FAILED) {
            fprintf(stderr, "mmap() failed.\n");
            return 1;
        }
    }
    for(i=0; i<video->nbuffers; i++) {
        struct v4l2_buffer buf;
        memset(&buf, 0, sizeof(buf));
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = i;
        if(ioctl(video->fd, VIDIOC_QBUF, &buf) < 0) {
            fprintf(stderr, "VIDIOC_QBUF ioctl failed.\n");
            return 1;
        }
    }
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(ioctl(video->fd, VIDIOC_STREAMON, &type) < 0) {
        fprintf(stderr, "VIDIOC_STREAMON ioctl failed.\n");
        return 1;
    }

    return 0;
}

int video_close(Video *video) {
    enum v4l2_buf_type type;
    int i;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(ioctl(video->fd, VIDIOC_STREAMOFF, &type) < 0) {
        fprintf(stderr, "VIDIOC_STREAMOFF ioctl failed.\n");
        return 1;
    }
    for(i=0; i<video->nbuffers; i++) {
        munmap(video->buffers[i].start, video->buffers[i].length);
    }
    free(video->buffers);
    close(video->fd);
    return 0;
}

Grey *video_grey_snapshot(Video *video) {
    struct v4l2_buffer buf;
    int i;
    Grey *grey;
    unsigned char *data;
    fd_set fds;
    struct timeval tv;
    int r;

tryagain:;
    FD_ZERO(&fds);
    FD_SET(video->fd, &fds);
    tv.tv_sec = 1;
    tv.tv_usec = 0;
    r = select(video->fd+1, &fds, NULL, NULL, &tv);
    if(r < 0) {
        if(errno == EINTR) goto tryagain;
        return NULL;
    }
    if(r == 0) {
        return NULL;
    }
    memset(&buf, 0, sizeof(buf));
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if(ioctl(video->fd, VIDIOC_DQBUF, &buf) < 0) {
        switch(errno) {
            case EAGAIN:
                goto tryagain;
            default:
                return NULL;
        }
    }
    data = video->buffers[buf.index].start;
    grey = malloc(sizeof(*grey)+video->x*video->y*sizeof(*(grey->pixels)));
    grey->x = video->x;
    grey->y = video->y;
    for(i=0; i<video->x*video->y; i++) {
        grey->pixels[i] = (float)data[2*i] / 255.0f;
    }
    ioctl (video->fd, VIDIOC_QBUF, &buf);
    return grey;
}

